using System;
using System.Collections.ObjectModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Threading;
using Infragistics.Windows.DataPresenter;
using Infragistics.Windows.DataPresenter.Events;



namespace SSI.Windows.InfragisticsExtensions
{
	/// <summary>
	/// Class for connecting a field column to a framework element
	/// </summary>
	public class FieldAdorner
	{
		/// <summary>
		/// Gets/sets the name of the field the adorner is attached to
		/// </summary>
		public String Name { get; set; }

		/// <summary>
		/// Gets/sets the adorning control
		/// </summary>
		public FrameworkElement Adorner { get; set; }
	}

	/// <summary>
	/// Collection of FieldAdorner
	/// </summary>
	public class FieldAdornerCollection : ObservableCollection<FieldAdorner>
	{
		/// <summary>
		/// Gets the adorner from the field name
		/// </summary>
		/// <param name="fieldname"></param>
		/// <returns>The adorner associated with the field name</returns>
		public FieldAdorner this[String fieldname]
		{
			get
			{
				foreach (var item in this.Items)
				{
					if (item.Name == fieldname)
						return (item);
				}

				return (null);
			}
		}

		/// <summary>
		/// Returns true if the field name is found in the collection
		/// </summary>
		/// <param name="fieldname">The field name</param>
		/// <returns>True if the collection contains the field name</returns>
		public Boolean ContainsField (String fieldname)
		{
			foreach (var item in this.Items)
			{
				if (item.Name == fieldname)
					return (true);
			}

			return (false);
		}
	}

	/// <summary>
	/// Class for FieldAdorner settings to be used with a XamDataGrid
	/// </summary>
	public class FieldAdornerSettings
	{
		#region Construction
		/// <summary>
		/// Default constructor
		/// </summary>
		public FieldAdornerSettings ()
		{
		}
		#endregion

		#region Members
		private DataPresenterBase presenter = null;
		private UIElementAdorner<Control> adorner = null;
		#endregion

		#region Properties
		public FieldAdornerCollection FieldAdorners { get; set; } = new FieldAdornerCollection();

		/// <summary>
		/// Gets whether the adorner is visible
		/// </summary>
		protected Boolean IsAdornerVisible
		{
			get
			{
				return (this.adorner.Visibility == Visibility.Visible);
			}
		}
		#endregion

		#region Interface
		/// <summary>
		/// Attaches the adorner to the presenter
		/// </summary>
		/// <param name="presenter">The presenter</param>
		public void Attach (DataPresenterBase presenter)
		{
			this.presenter = presenter;
			this.adorner = new UIElementAdorner<Control>(this.presenter);
			this.HookEventHandlers();
		}

		/// <summary>
		/// Detaches adorners
		/// </summary>
		public void Detach ()
		{
			this.UnhookEventHandlers();
		}
		#endregion

		#region Event Handlers
		// This method is invoked when the value of a cell is modified.
		private void CellChanged (Object sender, CellChangedEventArgs e)
		{
			this.VerifyAdorner();
		}

		private void CellActivated (Object sender, CellActivatedEventArgs e)
		{
			this.VerifyAdorner();
		}

		private void EditModeEnded (Object sender, EditModeEndedEventArgs e)
		{
			this.VerifyAdorner();
		}

		private void EditModeStarted (Object sender, EditModeStartedEventArgs e)
		{
			this.VerifyAdorner();
		}

		private void GroupByAreaExpansionStateChanged (Object sender, RoutedEventArgs e)
		{
			this.MoveAdornerWhileGroupByAreaExpansionStateChanges();
		}

		private void Grouped (Object sender, GroupedEventArgs e)
		{
			this.VerifyAdornerAsync();
		}

		private void IsKeyboardFocusWithinChanged (Object sender, DependencyPropertyChangedEventArgs e)
		{
			// This is necessary for when the grid loses focus, and then 
			// gets focus back again.  We need to see if the adorning
			// editor should be displayed for the active cell.
			if (this.presenter.IsKeyboardFocusWithin && this.presenter.ActiveCell != null && Keyboard.FocusedElement == this.presenter)
				this.VerifyAdorner();
		}

		private void PreviewKeyDown (Object sender, KeyEventArgs e)
		{
			e.Handled = this.ProcessKeyStroke(e.Key, e.KeyboardDevice);
		}

		private void RecordActivated (Object sender, RecordActivatedEventArgs e)
		{
			this.VerifyAdorner();
		}

		private void RecordAdded (Object sender, RecordAddedEventArgs e)
		{
			this.VerifyAdornerAsync();
		}

		private void RecordCollapsed (Object sender, RecordCollapsedEventArgs e)
		{
			this.VerifyAdornerAsync();
		}

		private void RecordExpanded (Object sender, RecordExpandedEventArgs e)
		{
			this.VerifyAdornerAsync();
		}

		private void RecordsInViewChanged (Object sender, RecordsInViewChangedEventArgs e)
		{
			this.VerifyAdorner();
		}

		private void SizeChanged (Object sender, SizeChangedEventArgs e)
		{
			this.RelocateAdorner();
		}

		private void Sorted (Object sender, SortedEventArgs e)
		{
			this.presenter.Dispatcher.BeginInvoke(DispatcherPriority.Background, (Action) (() =>
			{
				if (this.adorner.Child != null)
					this.RelocateAdorner();
				else
					this.VerifyAdorner();
			}));
		}
		#endregion

		#region Private Helpers
		private void HookEventHandlers ()
		{
			this.presenter.CellChanged += this.CellChanged;
			this.presenter.CellActivated += this.CellActivated;
			this.presenter.EditModeEnded += this.EditModeEnded;
			this.presenter.EditModeStarted += this.EditModeStarted;
			this.presenter.GroupByArea.Collapsed += this.GroupByAreaExpansionStateChanged;
			this.presenter.GroupByArea.Expanded += this.GroupByAreaExpansionStateChanged;
			this.presenter.Grouped += this.Grouped;
			this.presenter.IsKeyboardFocusWithinChanged += this.IsKeyboardFocusWithinChanged;
			this.presenter.PreviewKeyDown += this.PreviewKeyDown;
			this.presenter.RecordActivated += this.RecordActivated;
			this.presenter.RecordAdded += this.RecordAdded;
			this.presenter.RecordCollapsed += this.RecordCollapsed;
			this.presenter.RecordExpanded += this.RecordExpanded;
			this.presenter.RecordsInViewChanged += this.RecordsInViewChanged;
			this.presenter.SizeChanged += this.SizeChanged;
			this.presenter.Sorted += this.Sorted;
		}

		private void UnhookEventHandlers ()
		{
			this.presenter.CellChanged -= this.CellChanged;
			this.presenter.CellActivated -= this.CellActivated;
			this.presenter.EditModeEnded -= this.EditModeEnded;
			this.presenter.EditModeStarted -= this.EditModeStarted;
			this.presenter.GroupByArea.Collapsed -= this.GroupByAreaExpansionStateChanged;
			this.presenter.GroupByArea.Expanded -= this.GroupByAreaExpansionStateChanged;
			this.presenter.Grouped -= this.Grouped;
			this.presenter.IsKeyboardFocusWithinChanged -= this.IsKeyboardFocusWithinChanged;
			this.presenter.PreviewKeyDown -= this.PreviewKeyDown;
			this.presenter.RecordActivated -= this.RecordActivated;
			this.presenter.RecordAdded -= this.RecordAdded;
			this.presenter.RecordCollapsed -= this.RecordCollapsed;
			this.presenter.RecordExpanded -= this.RecordExpanded;
			this.presenter.RecordsInViewChanged -= this.RecordsInViewChanged;
			this.presenter.SizeChanged -= this.SizeChanged;
			this.presenter.Sorted -= this.Sorted;
		}

		private void GiveFocusToEditor ()
		{
			if (this.adorner.Child == null || !this.adorner.Child.IsTabStop)
				return;

			// Setting focus to the adorner first ensures that the controls
			// inside the editor can receive focus properly.
			FocusManager.SetFocusedElement(this.presenter, this.adorner);
			FocusManager.SetFocusedElement(this.presenter, this.adorner.Child);
		}

		private void HideAdorner ()
		{
			if (this.IsAdornerVisible)
			{
				this.adorner.Visibility = Visibility.Collapsed;
				this.adorner.Child = null;
			}
		}

		#region MoveAdornerToActiveCell
		private Boolean MoveAdornerToActiveCell (Control adorningeditor)
		{
			if (this.presenter.ActiveCell == null)
				return (false);

			if (adorningeditor != null && this.adorner.Child != adorningeditor)
				this.adorner.Child = adorningeditor;

			CellValuePresenter cvp = this.GetCellValuePresenterForActiveCell();
			if (cvp != null)
			{
				Point location = this.CalculateAdornerLocation(cvp);
				this.adorner.SetOffsets(location.X, location.Y);
				return (true);
			}

			return (false);
		}

		private CellValuePresenter GetCellValuePresenterForActiveCell ()
		{
			Cell activecell = this.presenter.ActiveCell;
			if (activecell == null)
				return (null);

			CellValuePresenter cvp = CellValuePresenter.FromCell(activecell);
			if (cvp == null || !cvp.IsDescendantOf(this.presenter) || !cvp.IsVisible)
				return (null);

			return (cvp);
		}

		private Point CalculateAdornerLocation (CellValuePresenter cvp)
		{
			GeneralTransform trans = cvp.TransformToAncestor(this.presenter);
			Point location = trans.Transform(new Point(0, 0));

			this.VerifyAdornerMetrics();

			IScrollInfo scrollinfo = this.presenter.ScrollInfo;

			// Make sure the adorner does not extend past the grid's bottom edge.
			Boolean isscrollinghorizontally = scrollinfo.ViewportWidth < scrollinfo.ExtentWidth;
			Double horizontalscrollbarheight = isscrollinghorizontally ? SystemParameters.ScrollHeight : 0;

			Double bottomedgeofadorner = location.Y + cvp.ActualHeight + this.adorner.ActualHeight;
			if ((this.presenter.ActualHeight - horizontalscrollbarheight) < bottomedgeofadorner)
				location.Offset(0, -this.adorner.ActualHeight);
			else
				location.Offset(0, cvp.ActualHeight);

			// Make sure the adorner does not extend past the grid's right edge.
			Boolean isscrollingvertically = scrollinfo.ViewportHeight < scrollinfo.ExtentHeight;
			Double verticalscrollbarwidth = isscrollingvertically ? SystemParameters.ScrollWidth : 0;

			Double rightedgeofadorner = location.X + this.adorner.ActualWidth;
			Double horizontaldiff = rightedgeofadorner - (this.presenter.ActualWidth - verticalscrollbarwidth);
			if (0 < horizontaldiff)
				location.Offset(-horizontaldiff, 0);

			return (location);
		}

		private void VerifyAdornerMetrics ()
		{
			this.VerifyAdornerLayer();

			if (!this.IsAdornerVisible)
				this.adorner.Visibility = Visibility.Visible;

			// Force the adorner through a measure pass since the hosted editor control
			// might have changed since the last time the metrics were calculated.
			this.adorner.InvalidateMeasure();
			this.adorner.UpdateLayout();
		}
		#endregion

		private void MoveAdornerWhileGroupByAreaExpansionStateChanges ()
		{
			if (!this.IsAdornerVisible)
				return;

			DispatcherTimer timer = new DispatcherTimer() { Interval = TimeSpan.FromMilliseconds(1) };
			Int32 counter = 0;
			timer.Tick += delegate
			{
				if (++counter > 500)
				{
					timer.Stop();
					timer = null;
				}
				else if (this.IsAdornerVisible)
				{
					this.MoveAdornerToActiveCell(null);
				}
			};
			timer.Start();
		}

		private Boolean ProcessKeyStroke (Key key, KeyboardDevice device)
		{
			if (!this.IsAdornerVisible)
				return (false);

			Boolean giveeditorfocus = key == Key.Tab && device.Modifiers != ModifierKeys.Shift && this.presenter.ActiveCell != null && this.adorner.Child != null && this.adorner.Child.IsTabStop;
			if (!giveeditorfocus)
				return (false);

			this.GiveFocusToEditor();

			return (true);
		}

		private void RelocateAdorner ()
		{
			if (this.IsAdornerVisible)
			{
				Boolean resumeediting = false;

				if (this.adorner.IsKeyboardFocusWithin)
				{
					resumeediting = true;

					// We must take the adorner out of focus otherwise it won't move.
					this.presenter.Focus();
				}

				// Delay the call that moves the adorner so that the
				// grid has a chance to finish updating its elements.
				this.presenter.Dispatcher.BeginInvoke(
					DispatcherPriority.Background,
					(Action) delegate
					{
						this.MoveAdornerToActiveCell(null);

						if (resumeediting)
							this.GiveFocusToEditor();
					});
			}
		}

		private void VerifyAdorner ()
		{
			//if the cell is a filter cell don't do anything with the adorner
			if (this.presenter.ActiveCell is FilterCell)
				return;

			this.VerifyAdornerLayer();

			FrameworkElement adorningeditor = null;
			if (this.presenter.ActiveCell != null)
			{
				var fieldadorner = this.FieldAdorners[this.presenter.ActiveCell.Field.Name];
				if (fieldadorner != null)
				{
					adorningeditor = fieldadorner.Adorner;
					if (adorningeditor is IXamDataPresenterFieldAdorner presenterfieldadorner && !presenterfieldadorner.IsAttached)
						presenterfieldadorner.Attach(this.presenter);

					adorningeditor.DataContext = this.presenter.ActiveCell.Record.DataItem;
				}
			}

			Boolean movedadorner = false;
			if (adorningeditor != null)
				movedadorner = this.MoveAdornerToActiveCell((Control) adorningeditor);

			if (!movedadorner)
				this.HideAdorner();
		}

		private void VerifyAdornerAsync ()
		{
			this.presenter.Dispatcher.BeginInvoke(DispatcherPriority.Background, (Action) (() => this.VerifyAdorner()));
		}

		/// <summary>
		/// Ensures that the adorner is in the adorner layer.
		/// </summary>
		/// <returns>True if the adorner is in the adorner layer, else false.</returns>
		private Boolean VerifyAdornerLayer ()
		{
			if (this.adorner.Parent != null)
				return (true);

			AdornerLayer layer = AdornerLayer.GetAdornerLayer(this.presenter);
			if (layer == null)
				return (false);

			layer.Add(this.adorner);
			return (true);
		}
		#endregion
	}
}