﻿using System;
using System.Collections;
using System.Windows;
using System.Windows.Documents;
using System.Windows.Media;



namespace SSI.Windows.InfragisticsExtensions
{
	/// <summary>
	/// An adorner that can display one and only one UIElement.  
	/// That element can be a panel, which contains multiple other elements.
	/// The element is added to the adorner's visual and logical trees, enabling it to 
	/// particpate in dependency property value inheritance, amongst other things.
	/// </summary>
	public class UIElementAdorner<T> : Adorner where T : UIElement
	{
		#region Construction
		/// <summary>
		/// Default Constructor
		/// </summary>
		/// <param name="adornedelement">The element to which the adorner will be bound.</param>
		public UIElementAdorner (UIElement adornedelement)
			: base(adornedelement)
		{
		}
		#endregion

		#region Child
		private T child = null;

		/// <summary>
		/// Gets/sets the child element hosted in the adorner.
		/// </summary>
		public T Child
		{
			get
			{
				return (this.child);
			}
			set
			{
				if (value == this.child)
					return;

				if (this.child != null)
				{
					base.RemoveLogicalChild(this.child);
					base.RemoveVisualChild(this.child);
				}

				this.child = value;

				if (this.child != null)
				{
					base.AddLogicalChild(this.child);
					base.AddVisualChild(this.child);
				}
			}
		}
		#endregion

		#region OffsetLeft
		private Double offsetleft = 0;

		/// <summary>
		/// Gets/sets the horizontal offset of the adorner.
		/// </summary>
		public Double OffsetLeft
		{
			get
			{
				return (this.offsetleft);
			}
			set
			{
				this.offsetleft = value;
				this.UpdateLocation();
			}
		}
		#endregion

		#region OffsetTop
		private Double offsettop = 0;

		/// <summary>
		/// Gets/sets the vertical offset of the adorner.
		/// </summary>
		public Double OffsetTop
		{
			get
			{
				return (this.offsettop);
			}
			set
			{
				this.offsettop = value;
				this.UpdateLocation();
			}
		}
		#endregion

		#region Interface
		/// <summary>
		/// Gets the desired transform
		/// </summary>
		/// <param name="transform">The transform</param>
		/// <returns>The transform</returns>
		public override GeneralTransform GetDesiredTransform (GeneralTransform transform)
		{
			GeneralTransformGroup result = new GeneralTransformGroup();
			result.Children.Add(base.GetDesiredTransform(transform));
			result.Children.Add(new TranslateTransform(this.OffsetLeft, this.OffsetTop));
			return (result);
		}

		/// <summary>
		/// Updates the location of the adorner in one atomic operation.
		/// </summary>
		public void SetOffsets (Double left, Double top)
		{
			this.offsetleft = left;
			this.offsettop = top;
			this.UpdateLocation();
		}
		#endregion

		#region Protected Overrides
		protected override Size MeasureOverride (Size constraint)
		{
			if (this.Child == null)
				return (base.MeasureOverride(constraint));

			this.Child.Measure(constraint);
			return (this.Child.DesiredSize);
		}

		protected override Size ArrangeOverride (Size finalsize)
		{
			if (this.Child == null)
				return (base.ArrangeOverride(finalsize));

			this.Child.Arrange(new Rect(finalsize));
			return (finalsize);
		}

		protected override IEnumerator LogicalChildren
		{
			get
			{
				ArrayList list = new ArrayList();
				if (this.Child != null)
					list.Add(this.Child);
				return (list.GetEnumerator());
			}
		}

		protected override Int32 VisualChildrenCount
		{
			get
			{
				return (this.Child == null ? 0 : 1);
			}
		}

		protected override Visual GetVisualChild (Int32 index)
		{
			return (this.Child);
		}
		#endregion

		#region Private Helpers
		void UpdateLocation ()
		{
			AdornerLayer adornerlayer = this.Parent as AdornerLayer;
			adornerlayer?.Update(this.AdornedElement);
		}
		#endregion
	}
}