﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using SSI.Analysis.Simulation.Expressions;



namespace SSI.Windows.Simulation.Conjoint
{
	internal static class FormulaHelpers
	{
		public readonly static String[] HelperTexts = new String[]
			{
				"=RANGE(start, end, [increment], [defaultvalue])",
				"=DEFAULTVALUE(value)"
			};

		public readonly static String NoMatchText = null;

		/// <summary>
		/// Match the formula helper text according to currently typed text
		/// </summary>
		/// <param name="helpertexts">The formula prototypes</param>
		/// <param name="typedtext">The typed in data from the user</param>
		/// <param name="nomatchtext">The text to return when no matches are found</param>
		/// <param name="valuemappings">Mappings of labels to values to use in parsing</param>
		/// <param name="locale">The locale for parsing parameters</param>
		/// <returns>A collection of strings containing the pre-highlighted, highlighted and post-highlighted sections</returns>
		public static String[] GetHelperHighlighting (IEnumerable<String> helpertexts, String typedtext, String nomatchtext, IDictionary<String, Double> valuemappings, CultureInfo locale)
		{
			if (String.IsNullOrWhiteSpace(typedtext))
				return (new String[] { nomatchtext, null, null });

			Char separator = locale.TextInfo.ListSeparator[0];

			String pre = null, highlight = null, post = null;

			if (typedtext == "=") // match all the formulas
				pre = String.Join("\n", helpertexts);
			else if (!typedtext.Contains("(")) // they're only up to the first paren, so match on what's typed so far
			{
				var matches = helpertexts.Where(t => t.StartsWith(typedtext, StringComparison.InvariantCultureIgnoreCase)).ToArray();
				if (matches != null && matches.Length > 0)
					pre = String.Join("\n", matches);
				else
					pre = nomatchtext;
			}
			else if (typedtext.EndsWith("(") && typedtext.Count(c => c == '(') == 1) // first token
			{
				var matches = helpertexts.Where(t => t.StartsWith(typedtext, StringComparison.InvariantCultureIgnoreCase)).ToArray();
				if (matches == null || matches.Length == 0)
					pre = nomatchtext;
				else if (matches.Length > 1)
					pre = String.Join("\n", matches);
				else
				{
					String match = matches.Single();
					Int32 typedtextlength = typedtext.Length;

					pre = match.Substring(0, typedtextlength);

					Int32 highlightendidx = match.IndexOfAny(new Char[] { separator, ')' }, typedtextlength);
					highlight = match.Substring(typedtextlength, highlightendidx - typedtextlength);
					post = match.Substring(highlightendidx);
				}
			}
			else // now the crud hits the fan
			{
				Int32 firstopenparen = typedtext.IndexOf('(');
				String functionname = typedtext.Substring(0, firstopenparen);
				var argumentidx = ScenarioSpecExpressionParser.ParseSyntaxHighlighting(typedtext, valuemappings, locale);

				var matches = helpertexts.Where(t => t.StartsWith(functionname, StringComparison.InvariantCultureIgnoreCase)).ToArray();
				if (matches == null || matches.Length == 0)
					pre = nomatchtext;
				else if (matches.Length > 1)
					pre = String.Join("\n", matches);
				else
				{
					String match = matches.Single();

					var matchtokens = match.Tokenize('(', separator, ')').AsArray();
					if (argumentidx == -1)
						pre = match;
					else
					{
						highlight = matchtokens[argumentidx + 1];
						Int32 highlightidx = match.IndexOf(highlight);
						pre = match.Substring(0, highlightidx);
						post = match.Substring(highlightidx + highlight.Length);
					}
				}
			}

			return (new String[] { pre, highlight, post });
		}
	}
}